#include <YoggieEngine.h>
#include "Physics.h"

namespace YoggieEngine {

	Physics::Physics()
	{
		// startup PhysX
		mFoundation = PxCreateFoundation(PX_PHYSICS_VERSION, gDefaultAllocatorCallback, gDefaultErrorCallback);
		if (!mFoundation) {
			spdlog::critical("PxCreateFoundation failed!");
		}

		mPvd = physx::PxCreatePvd(*mFoundation);
		if (!mPvd) {
			spdlog::critical("pxCreatPvd failed!");
		}
		transport = PxDefaultPvdSocketTransportCreate(PVD_HOST, 5425, 10);
		mPvd->connect(*transport, PxPvdInstrumentationFlag::eALL);


		mPhysics = PxCreatePhysics(PX_PHYSICS_VERSION, *mFoundation, PxTolerancesScale(), recordMemoryAllocations, mPvd);
		if (!mPhysics) {
			spdlog::critical("pxCreatePhysics failed!");
		}
		gDispatcher = nullptr;
		gMaterial = nullptr;
		mScene = nullptr;


	}

	Physics::~Physics()
	{
		if (mScene)
			mScene->release();

		if (gDispatcher)
			gDispatcher->release();

		if (mPhysics)
			mPhysics->release();

		if (mPvd) {
			if (mPvd->isConnected()) {
				mPvd->disconnect();
			}
			mPvd->release();
		}

		if (transport)
			transport->release();


		if (mFoundation)
			mFoundation->release();
	}

	void Physics::Step(float dt)
	{
		mScene->simulate(dt);
		mScene->fetchResults(true);
	}

	void Physics::Demo()
	{
		createScene();

		SetupPvdDebug();

		createGroundPlane();

		createStack(PxTransform(PxVec3(0, 0, stackZ -= 10.0f)), 10, 2.0f);
	}

	PxRigidDynamic* Physics::createDynamic(const PxTransform& t, const PxGeometry& geometry, const PxVec3& velocity = PxVec3(0)) {
		PxRigidDynamic* dynamic = PxCreateDynamic(*mPhysics, t, geometry, *gMaterial, 10.0f);
		dynamic->setAngularDamping(0.5f);
		dynamic->setLinearVelocity(velocity);
		mScene->addActor(*dynamic);
		return dynamic;
	}

	void Physics::createStack(const PxTransform& t, PxU32 size, PxReal halfextent)
	{
		PxShape* shape = mPhysics->createShape(PxBoxGeometry(halfextent, halfextent, halfextent), *gMaterial);
		for (PxU32 i = 0; i < size; i++) {
			for (PxU32 j = 0; j < size - i; j++) {
				PxTransform localTm(PxVec3(PxReal(j * 2) - PxReal(size - i), PxReal(i * 2 + 1), 0) * halfextent);
				PxRigidDynamic* body = mPhysics->createRigidDynamic(t.transform(localTm));
				body->attachShape(*shape);
				PxRigidBodyExt::updateMassAndInertia(*body, 10.0f);
				mScene->addActor(*body);
			}
		}

		shape->release();

	}

	void Physics::createScene()
	{
		PxSceneDesc sceneDesc(mPhysics->getTolerancesScale());
		sceneDesc.gravity = PxVec3(0.0f, -9.81f, 0.0f);
		gDispatcher = PxDefaultCpuDispatcherCreate(2);
		sceneDesc.cpuDispatcher = gDispatcher;
		sceneDesc.filterShader = PxDefaultSimulationFilterShader;
		mScene = mPhysics->createScene(sceneDesc);

	}

	void Physics::createGroundPlane()
	{
		gMaterial = mPhysics->createMaterial(0.5f, 0.5f, 0.6f);
		PxRigidStatic* groundPlane = PxCreatePlane(*mPhysics, PxPlane(0,1,0,0), *gMaterial);
		mScene->addActor(*groundPlane);

	}

	void Physics::SetupPvdDebug()
	{
		PxPvdSceneClient* pvdClient = mScene->getScenePvdClient();
		if (pvdClient) {
			pvdClient->setScenePvdFlag(PxPvdSceneFlag::eTRANSMIT_CONSTRAINTS, true);
			pvdClient->setScenePvdFlag(PxPvdSceneFlag::eTRANSMIT_CONTACTS, true);
			pvdClient->setScenePvdFlag(PxPvdSceneFlag::eTRANSMIT_SCENEQUERIES, true);
		}


	}

}